"""Utilities for validating against DLHub schemas"""
from typing import Union

from jsonschema import Draft7Validator, RefResolver
import requests

from dlhub_sdk.models import BaseMetadataModel

_schema_repo = "https://raw.githubusercontent.com/DLHub-Argonne/dlhub_schemas/master/schemas/"


def validate_against_dlhub_schema(document: Union[dict, BaseMetadataModel], schema_name: str):
    """Validate a metadata document against one of the DLHub schemas

    Note: Requires an internet connection

    Args:
        document: Document instance to be validated
        schema_name (string): Name of schema (e.g., "dataset" for validating datasets).
            For full list, see: https://github.com/DLHub-Argonne/dlhub_schemas
    Raises:
        (jsonschema.SchemaError) If the schema fails to validate
    """

    # Convert to dictionary, if needed
    if isinstance(document, BaseMetadataModel):
        document = document.to_dict()

    # Make the schema validator
    schema = requests.get("{}/{}.json".format(_schema_repo, schema_name)).json()
    validator = Draft7Validator(schema, resolver=RefResolver(_schema_repo, schema))

    # Test the document
    validator.validate(document)



"""This module contains tools for describing objects being published to DLHub."""
import importlib
from typing import List, Sequence, Optional, Iterable, Union, Dict
import pkg_resources
import requests
from enum import Enum

from pydantic import BaseModel, Field
from six import string_types
from zipfile import ZipFile
from glob import glob
import fnmatch
import json
import sys
import os
import re

from dlhub_sdk.models.datacite import Datacite, DataciteRelatedIdentifierType, DataciteRelationType
from dlhub_sdk.version import __version__

name_re = re.compile(r'^\S+$')


class DLHubType(Enum):
    """Type supported by DLHub"""

    servable = "servable"
    dataset = "dataset"
    pipeline = "piepline"


class DLHubMetadata(BaseModel):
    """Basic metadata for a DLHub artefact

    Includes information used by the DLHub web service to recognize an object,
    define how to build the computational environment,
    control access to it, et cetera"""

    version: str = Field(__version__, description="Version of the SDK used to create this object")
    domains: List[str] = Field(default_factory=list, description="List of scientific domains that describe this object")
    visible_to: List[str] = Field(['public'],
                                  description="List of groups or identities for which this object is visible")
    name: Optional[str] = Field(None, description="Name of the object", regex=r'^\S+$')
    files: Dict[str, Union[str, List[str]]] = Field(default_factory=dict,
                                                    description="List of files associated with this object")
    dependencies: Dict[str, Dict[str, str]] = Field(default_factory=dict,
                                                    description="Library dependencies for an object")
    test: Optional[bool] = Field(None, description="Internal use only. Used for testing the DLHub service")
    type: DLHubType = Field(DLHubType.servable, description="Type of this object")


class BaseMetadataModel(BaseModel):
    """Base class for models describing objects published via DLHub

    Covers information that goes in the :code:`datacite` block of the metadata file and
    some of the DLHub block.

    Using a MetadataModel
    ---------------------

    There are many kinds of MetadataModel classes that each describe a different kind of object.
    Each of these different types are created using the :code:`create_model` operation
    (e.g., :code:`KerasModel.create_model('model.h5')`), but have different arguments depending
    on the type of object. For example, TensorFlow models only require the directory created
    when saving the model for serving but scikit-learn models require the pickle file, how
    the pickle was created (e.g., with joblib), and how many input features it requires.

    Once created, you will need to fill in additional details about the object to make it reusable.
    The MetadataModel classes attempt to learn as much about an object as possible automatically,
    but there is some information that must be provided by a human. To start, you must define a
    title and name for the object and are encouraged to provide an abstract describing the model
    and list any associated papers/websites that describe the model. You will fill plenty of
    examples for how to describe the models in the DLHub_containers repostiory. Some types of
    objects require data specific to their type (e.g., Python servables need a list of required
    packages). We encourage you to find examples for your specific type of object in the containers
    repository for inspiration and to see the Python documentation for each Metadata Model.

    The MetadataModel object can be saved using the `to_dict` operation and read back into
    memory using the `from_dict` method. We recommend you save your dictionary to disk in the
    JSON or yaml format, which will allow for manual edits to be made before submitting or
    resubmitting a object description.
    """

    dlhub: DLHubMetadata = Field(default_factory=DLHubMetadata, description="Metadata used by the DLHub web service")
    datacite: Datacite = Field(default_factory=Datacite)

    @property
    def name(self) -> str:
        """
        Get the name of the servable

        Returns:
            (string) Name of the servable
        """
        return self.dlhub.name

    @classmethod
    def create_model(cls, **kwargs) -> 'BaseMetadataModel':
        """Instantiate the metadata model.

        Takes in arguments that allow metadata describing a dataset to be autogenerated. For
        example, these could include options describing how to read a dataset from a CSV file or
        which class method to invoke on a Python pickle object.
        """
        return cls()

    def read_codemeta_file(self, directory: Optional[str] = None):
        """Read in metadata from a codemeta.json file

        Args:
            directory (string): Path to directory contain the codemeta.json file (default: current working directory)
        """

        # If no directory, use the current
        if directory is None:
            directory = os.getcwd()

        # Load in the codemeta
        with open(os.path.join(directory, 'codemeta.json')) as fp:
            codemeta = json.load(fp)

        # Convert it to datacite and store it
        self.datacite = Datacite.from_codemeta(codemeta)

        return self

    def set_creators(self, authors: List[str], affiliations: List[Sequence[str]] = ()) -> 'BaseMetadataModel':
        """Add authors to this object

        Args:
            authors: List of authors for the dataset. In format: "<Family Name>, <Given Name>"
            affiliations: List of affiliations for each author.
        """
        self.datacite.set_creators(authors, affiliations)
        return self

    def set_title(self, title: str) -> 'BaseMetadataModel':
        """Set the title for this object

        Args:
            title: Desired title
        """
        self.datacite.set_title(title)
        return self

    def set_version(self, version):
        """Set the version of this resource

        Args:
            version (string): Version number
        """
        self.datacite.version = str(version)
        return self

    def set_abstract(self, abstract: str) -> 'BaseMetadataModel':
        """Define an abstract for this object. Use for a high-level summary

        Args:
            abstract: Description of this artifact
        """
        self.datacite.set_abstract(abstract)
        return self

    def set_methods(self, methods: str) -> 'BaseMetadataModel':
        """Define a methods section for this object. Use to describe any specific details
        about how the dataset, model, etc was generated.

        Args:
            methods (str): Detailed method descriptions
        """
        self.datacite.set_methods(methods)
        return self

    def set_domains(self, domains: List[str]):
        """Set the field of science that is associated with this object

        Args:
            domains: Names of fields of science (e.g., "materials science")
        """
        self.dlhub.domains = domains.copy()
        return self

    def set_visibility(self, users: Optional[Iterable[str]] = None,
                       groups: Optional[Iterable[str]] = None):
        """Set the list of people this object should be visible to.

        By default, it will be visible to anyone (["public"]).

        Args:
            users: GlobusAuth UUIDs of allowed users
            groups: GlobusAuth UUIDs of allowed Globus groups
        """
        visibilities = []
        if not (users or groups):
            visibilities.append('public')
        else:
            if users:
                for user in users:
                    visibilities.append('urn:globus:auth:identity:' + user)
            if groups:
                for group in groups:
                    visibilities.append('urn:globus:groups:id:' + group)
        self.dlhub.visible_to = visibilities
        return self

    def set_doi(self, doi):
        """Set the DOI of this object, if available

        This function is only for advanced usage. Most users of the toolbox will not
        know the DOI before sending the object in to DLHub.

        Args:
            doi (string): DOI of the object
        """
        self.datacite.set_identifier(doi)
        return self

    def set_name(self, name):
        """Set the name of the object.

        Should be something short, descriptive, and memorable

        Args:
            name (string): Name of artifact
        """
        if name_re.match(name) is None:
            raise ValueError('Name cannot contain any whitespace')
        self.dlhub.name = name
        return self

    def set_publication_year(self, year):
        """Define the publication year

        This function is only for advanced usage. Normally, this will be assigned automatically

        Args:
            year (string): Publication year
        """
        self.datacite.publicationYear = str(year)
        return self

    def add_requirement(self, library, version=None):
        """Add a required Python library.

        The name of the library should be either the name on PyPI, or a URL for the git repository
        holding the code (e.g., ``git+https://github.com/DLHub-Argonne/dlhub_sdk.git``)

        Args:
            library (string): Name of library
            version (string): Required version. 'latest' to use the most recent version on PyPi (if
                available). 'detect' will attempt to find the version of the library installed on
                the computer running this software. Default is ``None``
        """

        # Attempt to determine the version automatically
        if version == "detect":
            try:
                module = importlib.import_module(library)
                version = module.__version__
            except (AttributeError, ModuleNotFoundError):
                version = pkg_resources.get_distribution(library).version
        elif version == "latest":
            pypi_req = requests.get('https://pypi.org/pypi/{}/json'.format(library))
            version = pypi_req.json()['info']['version']

        # Set the requirements
        if "python" not in self.dlhub.dependencies:
            self.dlhub.dependencies["python"] = {}
        self.dlhub.dependencies["python"][library] = version
        return self

    def add_requirements(self, requirements):
        """Add several Python library requirements

        Args:
            requirements (dict): Keys are names of library (str), values are the version
        """
        for p, v in requirements.items():
            self.add_requirement(p, v)
        return self

    def parse_repo2docker_configuration(self, directory=None):
        """Gathers information about required environment from repo2docker configuration files.

        See https://repo2docker.readthedocs.io/en/latest/config_files.html for more details

        Args:
            directory (str): Path to directory containing configuration files
                (default: current working directory)
        """

        # Get a list of all files
        config_files = ['environment.yml', 'requirements.txt', 'setup.py', 'REQUIRE', 'install.R',
                        'apt.txt', 'DESCRIPTION', 'manifest.xml', 'postBuild', 'start',
                        'runtime.txt', 'default.nix', 'Dockerfile']

        # Get the directory name if `None`
        if directory is None:
            directory = os.getcwd()

        # Add every file we can find
        for file in config_files:
            path = os.path.join(directory, file)
            if os.path.isfile(path):
                self.add_file(path)

        return self

    def add_related_resource(self, identifier: str, identifier_type: Union[str, DataciteRelatedIdentifierType],
                             relation_type: Union[str, DataciteRelationType]) -> 'BaseMetadataModel':
        """Add a resource that is related to this object.

        We use the DataCite to describe the relations. Common relation types for DLHub objects are:
            - "IsDescribedBy": For a paper that describes a dataset or model
            - "IsDocumentedBy": For the software documentation for a model
            - "IsDerviedFrom": For the database a training set was pulled from
            - "Requires": For any software libraries that are required for this module

        Args:
            identifier: Identifier
            identifier_type: Identifier type
            relation_type: Relation between this identifier and the object you are describing
        """
        self.datacite.add_related_identifier(identifier, identifier_type, relation_type)

    def add_file(self, file, name=None):
        """Add a file to the list of files to be distributed with this object

        Args:
            file (string): Path to the file
            name (string): Optional. Name of the file, if it is a file that serves a specific
                purpose (e.g., "pickle" if this is a pickle file of a scikit-learn model)
        """

        if name is None or name == "other":
            if "other" not in self.dlhub.files:
                self.dlhub.files["other"] = []
            self.dlhub.files["other"].append(file)
        else:
            self.dlhub.files[name] = file
        return self

    def add_directory(self, directory: str, include: Union[str, Iterable[str]] = (),
                      exclude: Union[str, Iterable[str]] = (), recursive: bool = False):
        """Add all the files in a directory

        Args:
            include (string or [string]): Only add files that match any of these patterns
            exclude (string or [string]): Exclude all files that match any of these patterns
            directory (string): Path to a directory
            recursive (bool): Whether to add all files in a directory
        """

        # Make sure include and exclude are lists
        if isinstance(include, string_types):
            include = [include]
        if isinstance(exclude, string_types):
            exclude = [exclude]

        # Get potential files
        if recursive:
            if sys.version_info < (3, 5):
                raise RuntimeError('You must use Python >= 3.5 to do recursive listing')
            hits = glob('{}/**/*'.format(directory), recursive=True)
        else:
            hits = glob('{}/*'.format(directory))

        # Get only the files that match the filters
        files = [x for x in hits if os.path.isfile(x)]
        if len(include) > 0:  # Run inclusive filters
            files = [f for f in files
                     if any(fnmatch.fnmatch(os.path.basename(f), i) for i in include)]
        if len(exclude) > 0:
            files = [f for f in files
                     if not any(fnmatch.fnmatch(os.path.basename(f), e) for e in exclude)]

        return self.add_files(files)

    def add_files(self, files: Iterable[str]):
        """Add files that should be distributed with this artifact.

        Args:
            files: Paths of files that should be published
        """
        # Add the files
        for file in files:
            self.add_file(file)
        return self

    def to_dict(self, simplify_paths: bool = False):
        """Render the object to a JSON-ready dictionary

        Args:
            simplify_paths (bool): Whether to simplify the paths of each file
        Returns:
            (dict) A description of the dataset in a form suitable for download
        """

        # Check for required fields
        if len(self.datacite.titles) == 0:
            raise ValueError('Title must be specified. Use `set_title`')
        if self.name is None:
            raise ValueError('Name must be specified. Use `set_name`')

        # Make a copy of the output
        out = self.dict(exclude_none=True).copy()

        # Prepare the files
        if simplify_paths:
            # Get the common path
            common_path = self._get_common_path()

            files = {}
            for k, v in out["dlhub"]["files"].items():
                if k == "other":
                    files[k] = [os.path.relpath(f, common_path) for f in v]
                else:
                    files[k] = os.path.relpath(v, common_path)

            # Copy over the current files list
            out["dlhub"]["files"] = files
        return out

    def list_files(self):
        """Provide a list of files associated with this artifact.

        Returns:
            ([string]) list of file paths"""
        # Gather a list of all the files
        output = []
        for k, v in self.dlhub.files.items():
            if isinstance(v, string_types):
                output.append(v)
            else:  # It is a list
                output.extend(v)
        return output

    def get_zip_file(self, path):
        """Write all the listed files to a ZIP object

        Takes all of the files returned by `list_files`. First determines the largest common
        path of all files, and preserves directory structure by using this common path as the
        root directory. For example, if the files are "/home/a.pkl" and "/home/a/b.dat", the common
        directory is "/home" and the files will be stored in the Zip as "a.pkl" and "a/b.dat"

        Args:
            path (string): Path for the ZIP File
        Returns:
            (string): Base path for the ZIP file (useful for adjusting the paths of the files
                included in the metadata model)
        """

        # Open the zip file in "exclusively create" (x) mode
        with ZipFile(path, 'x') as newzip:
            if len(self.list_files()) == 0:
                return "."

            # Get the common path of all files
            root_path = self._get_common_path()

            # Add each file to the directory
            for file in self.list_files():
                newzip.write(file, arcname=os.path.relpath(file, root_path))

            return root_path

    def _get_common_path(self):
        """Determine the common path of all files

        Returns:
            (string) Common path
        """
        # Get the files
        files = self.list_files()

        # Shortcut: if no files
        if len(files) == 0:
            return '.'

        # Get the directories for all the files
        directories = [f if os.path.isdir(f) else os.path.dirname(f) for f in files]

        # Get the largest common path
        return os.path.commonpath(os.path.abspath(d) for d in directories)



import logging
import json
import os
from tempfile import mkstemp
from typing import Union, Any, Optional, Tuple, Dict

import requests
import globus_sdk
from globus_sdk.authorizers import GlobusAuthorizer
from globus_sdk.base import BaseClient, slash_join
from mdf_toolbox import login, logout
from mdf_toolbox.globus_search.search_helper import SEARCH_LIMIT
from funcx.sdk.client import FuncXClient

from dlhub_sdk.config import DLHUB_SERVICE_ADDRESS, CLIENT_ID
from dlhub_sdk.utils.futures import DLHubFuture
from dlhub_sdk.utils.schemas import validate_against_dlhub_schema
from dlhub_sdk.utils.search import DLHubSearchHelper, get_method_details, filter_latest

# Directory for authentication tokens
_token_dir = os.path.expanduser("~/.dlhub/credentials")
logger = logging.getLogger(__name__)


class DLHubClient(BaseClient):
    """Main class for interacting with the DLHub service

    Holds helper operations for performing common tasks with the DLHub service. For example,
    `get_servables` produces a list of all servables registered with DLHub.

    The initializer offers several routes for authenticating with the services:

    1. **Token-based authentication**. The default route for login is to log in with your own
        Globus-associated account. Provided no arguments, the DLHub client will ask you
        to authenticate with Globus Auth then store your login tokens on your computer.
        The DLHub SDK will re-use these tokens unless you later call :meth:`.logout`
        or initialize the client with ``force_login=True`.
    2. **Pre-defined authorizers**. The alternate route is to create authorizers
       `using the Globus SDK directly <https://globus-sdk-python.readthedocs.io/en/stable/tutorial/>`_
        and providing that authorizer to the initializer (e.g., ``DLHubClient(dlhub_authorizer=auth)``).
        You must provide authorizers DLHub for all sub-services: Globus Search, FuncX, and OpenID.
    """

    def __init__(self, dlh_authorizer: Optional[GlobusAuthorizer] = None,
                 search_authorizer: Optional[GlobusAuthorizer] = None,
                 fx_authorizer: Optional[GlobusAuthorizer] = None,
                 openid_authorizer: Optional[GlobusAuthorizer] = None,
                 http_timeout: Optional[int] = None,
                 force_login: bool = False, **kwargs):
        """Initialize the client

        Args:
            http_timeout (int): Timeout for any call to service in seconds. (default is no timeout)
            force_login (bool): Whether to force a login to get new credentials.
                A login will always occur if ``dlh_authorizer`` or ``search_client``
                are not provided.
            no_local_server (bool): Disable spinning up a local server to automatically
                copy-paste the auth code. THIS IS REQUIRED if you are on a remote server.
                When used locally with no_local_server=False, the domain is localhost with
                a randomly chosen open port number.
                **Default**: ``True``.
            no_browser (bool): Do not automatically open the browser for the Globus Auth URL.
                Display the URL instead and let the user navigate to that location manually.
                **Default**: ``True``.
            dlh_authorizer (:class:`GlobusAuthorizer <globus_sdk.authorizers.base.GlobusAuthorizer>`):
                An authorizer instance used to communicate with DLHub.
                If ``None``, will be created from your account's credentials.
            search_authorizer (:class:`GlobusAuthorizer <globus_sdk.authorizers.base.GlobusAuthorizer>`):
                An authenticated SearchClient to communicate with Globus Search.
                If ``None``, will be created from your account's credentials.
            fx_authorizer (:class:`GlobusAuthorizer
                            <globus_sdk.authorizers.base.GlobusAuthorizer>`):
                An authorizer instance used to communicate with funcX.
                If ``None``, will be created from your account's credentials.
            openid_authorizer (:class:`GlobusAuthorizer
                            <globus_sdk.authorizers.base.GlobusAuthorizer>`):
                An authorizer instance used to communicate with OpenID.
                If ``None``, will be created from your account's credentials.
        Keyword arguments are the same as for :class:`BaseClient <globus_sdk.base.BaseClient>`.
        """

        authorizers = [dlh_authorizer, search_authorizer, openid_authorizer, fx_authorizer]
        # Get authorizers through Globus login if any are not provided
        if not all(a is not None for a in authorizers):
            # If some but not all were provided, warn the user they could be making a mistake
            if any(a is not None for a in authorizers):
                logger.warning('You have defined some of the authorizers but not all. DLHub is falling back to login. '
                               'You must provide authorizers for DLHub, Search, OpenID, FuncX.')

            fx_scope = "https://auth.globus.org/scopes/facd7ccc-c5f4-42aa-916b-a0e270e2c2a9/all"
            auth_res = login(services=["search", "dlhub",
                                       fx_scope, "openid"],
                             app_name="DLHub_Client",
                             make_clients=False,
                             client_id=CLIENT_ID,
                             clear_old_tokens=force_login,
                             token_dir=_token_dir,
                             no_local_server=kwargs.get("no_local_server", True),
                             no_browser=kwargs.get("no_browser", True))

            # Unpack the authorizers
            dlh_authorizer = auth_res["dlhub"]
            fx_authorizer = auth_res[fx_scope]
            openid_authorizer = auth_res['openid']
            search_authorizer = auth_res['search']

        # Define the subclients needed by the service
        self._fx_client = FuncXClient(fx_authorizer=fx_authorizer,
                                      search_authorizer=search_authorizer,
                                      openid_authorizer=openid_authorizer,
                                      no_local_server=kwargs.get("no_local_server", True),
                                      no_browser=kwargs.get("no_browser", True))
        self._search_client = globus_sdk.SearchClient(authorizer=search_authorizer,
                                                      http_timeout=5 * 60)

        # funcX endpoint to use
        self.fx_endpoint = '86a47061-f3d9-44f0-90dc-56ddc642c000'
        self.fx_cache = {}
        super(DLHubClient, self).__init__("DLHub", environment='dlhub',
                                          authorizer=dlh_authorizer,
                                          http_timeout=http_timeout,
                                          base_url=DLHUB_SERVICE_ADDRESS,
                                          **kwargs)

    def logout(self):
        """Remove credentials from your local system"""
        logout()

    @property
    def query(self):
        """Access a query of the DLHub Search repository"""
        return DLHubSearchHelper(search_client=self._search_client)

    def get_username(self):
        """Get the username associated with the current credentials"""

        res = self.get('/namespaces')
        return res.data['namespace']

    def get_servables(self, only_latest_version=True):
        """Get all of the servables available in the service

        Args:
            only_latest_version (bool): Whether to only return the latest version of each servable
        Returns:
            ([list]) Complete metadata for all servables found in DLHub
        """

        # Get all of the servables
        results, info = self.query.match_field('dlhub.type', 'servable') \
            .add_sort('dlhub.owner', ascending=True).add_sort('dlhub.name', ascending=False) \
            .add_sort('dlhub.publication_date', ascending=False).search(info=True)
        if info['total_query_matches'] > SEARCH_LIMIT:
            raise RuntimeError('DLHub contains more servables than we can return in one entry. '
                               'DLHub SDK needs to be updated.')

        if only_latest_version:
            # Sort out only the most recent versions (they come first in the sorted list
            names = set()
            output = []
            for r in results:
                name = r['dlhub']['shorthand_name']
                if name not in names:
                    names.add(name)
                    output.append(r)
            results = output

        # Add these to the cache
        for r in results:
            self.fx_cache[r['dlhub']['shorthand_name']] = r['dlhub']['funcx_id']

        return results

    def list_servables(self):
        """Get a list of the servables available in the service

        Returns:
            [string]: List of all servable names in username/servable_name format
        """

        servables = self.get_servables(only_latest_version=True)
        return [x['dlhub']['shorthand_name'] for x in servables]

    def get_task_status(self, task_id):
        """Get the status of a DLHub task.

        Args:
            task_id (string): UUID of the task
        Returns:
            dict: status block containing "status" key.
        """

        r = self._fx_client.get_task(task_id)
        return r

    def describe_servable(self, name):
        """Get the description for a certain servable

        Args:
            name (string): DLHub name of the servable of the form <user>/<servable_name>
        Returns:
            dict: Summary of the servable
        """
        split_name = name.split('/')
        if len(split_name) < 2:
            raise AttributeError('Please enter name in the form <user>/<servable_name>')

        # Create a query for a single servable
        query = self.query.match_servable('/'.join(split_name[1:])) \
            .match_owner(split_name[0]).add_sort("dlhub.publication_date", False) \
            .search(limit=1)

        # Raise error if servable is not found
        if len(query) == 0:
            raise AttributeError('No such servable: {}'.format(name))
        return query[0]

    def describe_methods(self, name, method=None):
        """Get the description for the method(s) of a certain servable

        Args:
            name (string): DLHub name of the servable of the form <user>/<servable_name>
            method (string): Optional: Name of the method
        Returns:
             dict: Description of a certain method if ``method`` provided, all methods
                if the method name was not provided.
        """

        metadata = self.describe_servable(name)
        return get_method_details(metadata, method)

    def run(self, name: str, inputs: Any, parameters: Optional[Dict[str, Any]] = None,
            asynchronous: bool = False, debug: bool = False, async_wait: float = 5,
            timeout: Optional[float] = None)\
            -> Union[
                DLHubFuture,
                Tuple[Any, Dict[str, Any]],
                Any
            ]:
        """Invoke a DLHub servable

        Args:
            name: DLHub name of the servable of the form <user>/<servable_name>
            inputs: Data to be used as input to the function. Can be a string of file paths or URLs
            parameters: Any optional parameters to pass to the function.
            asynchronous: Whether to return from the function immediately or wait for the execution to finish.
            debug: Whether to capture the standard out and error printed during execution
            async_wait: How many seconds to wait between checking async status
            timeout: How long to wait for a result to return. Only used for synchronous calls
        Returns:
            If asynchronous, a DLHubFuture for the execution
            If debug, the output of the function and dictionary holding the following information:
                - success: Whether the code inside ran without raising an exception
                - stdout/stderr: Captured standard output and error, if requested
                - timing: Execution time for the segment in seconds
                - exc: Captured exception object
                - error_message: Exception traceback
            If neither, the output of the function
        """

        if name not in self.fx_cache:
            # Look it up and add it to the cache, this will raise an exception if not found.
            serv = self.describe_servable(name)
            self.fx_cache.update({name: serv['dlhub']['funcx_id']})

        funcx_id = self.fx_cache[name]
        payload = {
            'inputs': inputs,
            'parameters': parameters,
            'debug': debug
        }
        task_id = self._fx_client.run(payload, endpoint_id=self.fx_endpoint, function_id=funcx_id)

        # Return the result
        future = DLHubFuture(self, task_id, async_wait, debug)
        return future.result(timeout=timeout) if not asynchronous else future

    def run_serial(self, servables, inputs, async_wait=5):
        """Invoke each servable in a serial pipeline.
        This function accepts a list of servables and will run each one,
        passing the output of one as the input to the next.

        Args:
             servables (list): A list of servable strings
             inputs: Data to pass to the first servable
             asycn_wait (float): Seconds to wait between status checks
        Returns:
            Results of running the servable
        """
        if not isinstance(servables, list):
            print("run_serial requires a list of servables to invoke.")

        serv_data = inputs
        for serv in servables:
            serv_data = self.run(serv, serv_data, async_wait=async_wait)
        return serv_data

    def get_result(self, task_id, verbose=False):
        """Get the result of a task_id

        Args:
            task_id str: The task's uuid
            verbose bool: whether or not to return the full dlhub response
        Returns:
            Reult of running the servable
        """

        result = self._fx_client.get_result(task_id)
        if isinstance(result, tuple) and not verbose:
            result = result[0]
        return result

    def publish_servable(self, model):
        """Submit a servable to DLHub

        If this servable has not been published before, it will be assigned a unique identifier.

        If it has been published before (DLHub detects if it has an identifier), then DLHub
        will update the servable to the new version.

        Args:
            model (BaseMetadataModel): Servable to be submitted
        Returns:
            (string): Task ID of this submission, used for checking for success
        """

        # Get the metadata
        metadata = model.to_dict(simplify_paths=True)

        # Mark the method used to submit the model
        metadata['dlhub']['transfer_method'] = {'POST': 'file'}

        # Validate against the servable schema
        validate_against_dlhub_schema(metadata, 'servable')

        # Wipe the fx cache so we don't keep reusing an old servable
        self.clear_funcx_cache()

        # Get the data to be submitted as a ZIP file
        fp, zip_filename = mkstemp('.zip')
        os.close(fp)
        os.unlink(zip_filename)
        try:
            model.get_zip_file(zip_filename)

            # Get the authorization headers
            headers = {}
            self.authorizer.set_authorization_header(headers)

            # Submit data to DLHub service
            with open(zip_filename, 'rb') as zf:
                reply = requests.post(
                    slash_join(self.base_url, 'publish'),
                    headers=headers,
                    files={
                        'json': ('dlhub.json', json.dumps(metadata), 'application/json'),
                        'file': ('servable.zip', zf, 'application/octet-stream')
                    }
                )

            # Return the task id
            if reply.status_code != 200:
                raise Exception(reply.text)
            return reply.json()['task_id']
        finally:
            os.unlink(zip_filename)

    def publish_repository(self, repository):
        """Submit a repository to DLHub for publication

        Args:
            repository (string): Repository to publish
        Returns:
            (string): Task ID of this submission, used for checking for success
        """

        # Publish to DLHub
        metadata = {"repository": repository}

        # Wipe the fx cache so we don't keep reusing an old servable
        self.clear_funcx_cache()

        response = self.post('publish_repo', json_body=metadata)

        task_id = response.data['task_id']
        return task_id

    def search(self, query, advanced=False, limit=None, only_latest=True):
        """Query the DLHub servable library

        By default, the query is used as a simple plaintext search of all model metadata.
        Optionally, you can provided an advanced query on any of the indexed fields in
        the DLHub model metadata by setting :code:`advanced=True` and following the guide for
        constructing advanced queries found in the
        `Globus Search documentation <https://docs.globus.org/api/search/search/#query_syntax>`_.

        Args:
             query (string): Query to be performed
             advanced (bool): Whether to perform an advanced query
             limit (int): Maximum number of entries to return
             only_latest (bool): Whether to return only the latest version of the model
        Returns:
            ([dict]): All records matching the search query
        """

        results = self.query.search(query, advanced=advanced, limit=limit)
        return filter_latest(results) if only_latest else results

    def search_by_servable(self, servable_name=None, owner=None, version=None,
                           only_latest=True, limit=None, get_info=False):
        """Search by the ownership, name, or version of a servable

        Args:
            servable_name (str): The name of the servable. **Default**: None, to match
                    all servable names.
            owner (str): The name of the owner of the servable. **Default**: ``None``,
                    to match all owners.
            version (int): Model version, which corresponds to the date when the
                servable was published. **Default**: ``None``, to match all versions.
            only_latest (bool): When ``True``, will only return the latest version
                    of each servable. When ``False``, will return all matching versions.
                    **Default**: ``True``.
            limit (int): The maximum number of results to return.
                    **Default:** ``None``, for no limit.
            get_info (bool): If ``False``, search will return a list of the results.
                    If ``True``, search will return a tuple containing the results list
                    and other information about the query.
                    **Default:** ``False``.

        Returns:
            If ``info`` is ``False``, *list*: The search results.
            If ``info`` is ``True``, *tuple*: The search results,
            and a dictionary of query information.
        """
        if not servable_name and not owner and not version:
            raise ValueError("One of 'servable_name', 'owner', or 'publication_date' is required.")

        # Perform the query
        results, info = (self.query.match_servable(servable_name=servable_name, owner=owner,
                                                   publication_date=version)
                         .search(limit=limit, info=True))

        # Filter out the latest models
        if only_latest:
            results = filter_latest(results)

        if get_info:
            return results, info
        return results

    def search_by_authors(self, authors, match_all=True, limit=None, only_latest=True):
        """Execute a search for servables from certain authors.

        Authors in DLHub may be different than the owners of the servable and generally are
        the people who developed functionality of a certain servable (e.g., the creators
        of the machine learning model used in a servable).

        If you want to search by ownership, see :meth:`search_by_servable`

        Args:
            authors (str or list of str): The authors to match. Names must be in
                "Family Name, Given Name" format
            match_all (bool): If ``True``, will require all authors be on any results.
                    If ``False``, will only require one author to be in results.
                    **Default**: ``True``.
            limit (int): The maximum number of results to return.
                    **Default:** ``None``, for no limit.
            only_latest (bool): When ``True``, will only return the latest version
                    of each servable. When ``False``, will return all matching versions.
                    **Default**: ``True``.

        Returns:
            [dict]: List of servables from the desired authors
        """
        results = self.query.match_authors(authors, match_all=match_all).search(limit=limit)
        return filter_latest(results) if only_latest else results

    def search_by_related_doi(self, doi, limit=None, only_latest=True):
        """Get all of the servables associated with a certain publication

        Args:
            doi (string): DOI of related paper
            limit (int): Maximum number of results to return
            only_latest (bool): Whether to return only the most recent version of the model
        Returns:
            [dict]: List of servables from the requested paper
        """

        results = self.query.match_doi(doi).search(limit=limit)
        return filter_latest(results) if only_latest else results

    def clear_funcx_cache(self, servable=None):
        """Remove functions from the cache. Either remove a specific servable or wipe the whole cache.

        Args:
            Servable: str
                The name of the servable to remove. Default None
        """

        if servable:
            del (self.fx_cache[servable])
        else:
            self.fx_cache = {}

        return self.fx_cache



from itertools import zip_longest

from pydantic import AnyUrl, BaseModel, Field
from typing import List, Optional, Sequence, Union
from enum import Enum


class DataciteIdentifier(BaseModel):
    """A persistent identifier that identifies a resource. Currently, only DOI is allowed."""
    class DataciteIdentifierType(str, Enum):
        DOI = "DOI"

    identifier: str = '10.datacite/placeholder'
    identifierType: DataciteIdentifierType = DataciteIdentifierType.DOI


class DataciteCreator(BaseModel):
    """A researcher involved working on the data, or the authors of the publication in priority order.

    May be a corporate/institutional or personal name. Format: Family, Given"""

    class NameIdentifier(BaseModel):
        nameIdentifier: str
        nameIdentifierScheme: str
        schemeURI: Optional[AnyUrl]

    creatorName: Optional[str] = None
    nameIdentifiers: List[NameIdentifier] = Field(default_factory=list)
    affiliations: Optional[List[str]] = Field(default_factory=list)
    familyName: Optional[str] = None
    givenName: Optional[str] = None

    @classmethod
    def from_name(cls, name: str, affiliations: Optional[Sequence[str]]) -> 'DataciteCreator':
        """Create a Creator from a string in format "familyName, givenName"

        Args:
            name: Name of the creator
            affiliations: List of affiliations of the creator
        Returns:
            A DataciteCreator object
        """

        family_name, given_name = name.split(",", 1)
        return DataciteCreator(familyName=family_name.strip(), givenName=given_name.strip(),
                               affiliations=affiliations)


class DataciteTitle(BaseModel):
    class DataciteTitleType(Enum):
        AlternativeTitle = "AlternativeTitle"
        Subtitle = "Subtitle"
        TranslatedTitle = "TranslatedTitle"
        Other = "Other"

    title: str = Field(...)
    type: Optional[DataciteTitleType]
    lang: Optional[str]


class DataciteSubject(BaseModel):
    subject: str = ""
    subjectScheme: Optional[str] = ""
    schemeURI: Optional[AnyUrl] = ""
    valueURI: Optional[AnyUrl] = ""
    lang: str = ""


class DataciteResourceTypeGeneral(Enum):
    Audiovisual = "Audiovisual"
    Collection = "Collection"
    Dataset = "Dataset"
    Event = "Event"
    Image = "Image"
    InteractiveResource = "InteractiveResource"
    Model = "Model"
    PhysicalObject = "PhysicalObject"
    Service = "Service"
    Software = "Software"
    Sound = "Sound"
    Text = "Text"
    Workflow = "Workflow"
    Other = "Other"


class DataciteResourceType(BaseModel):
    resourceType: Optional[str] = ""
    resourceTypeGeneral: DataciteResourceTypeGeneral = None


class DataciteContributor(BaseModel):
    class DataciteContributorType(Enum):
        ContactPerson = "ContactPerson"
        DataCollector = "DataCollector"
        DataCurator = "DataCurator"
        DataManager = "DataManager"
        Editor = "Editor"
        HostingInstitution = "HostingInstitution"
        Other = "Other"
        Producer = "Producer"
        ProjectLeader = "ProjectLeader"
        ProjectManager = "ProjectManager"
        ProjectMember = "ProjectMember"
        RegistrationAgency = "RegistrationAgency"
        RegistrationAuthority = "RegistrationAuthority"
        RelatedPerson = "RelatedPerson"
        ResearchGroup = "ResearchGroup"
        RightsHolder = "RightsHolder"
        Researcher = "Researcher"
        Sponsor = "Sponsor"
        Supervisor = "Supervisor"
        WorkPackageLeader = "WorkPackageLeader"

    contributorType: DataciteContributorType = None
    contributorName: str = ""
    affiliations: Optional[List[str]] = Field(default_factory=list)
    familyName: Optional[str] = None
    givenName: Optional[str] = None


class DataciteAlternateIdentifier(BaseModel):
    """An identifier or identifiers other than the primary Identifier applied to the resource being registered.

     This may be any alphanumeric string which is unique within its domain of issue. May be used for local identifiers.
    AlternateIdentifier should be used for another identifier of the same instance (same location, same file)."""

    alternateIdentifier: str
    alternateIdentifierType: str


class DataciteRelatedIdentifierType(Enum):
    """Types of identifiers for different objects"""
    ARK = "ARK"
    arXiv = "arXiv"
    bibcode = "bibcode"
    DOI = "DOI"
    EAN13 = "EAN13"
    EISSN = "EISSN"
    Handle = "Handle"
    IGSN = "IGSN"
    ISBN = "ISBN"
    ISSN = "ISSN"
    ISTC = "ISTC"
    LISSN = "LISSN"
    LSID = "LSID"
    PMID = "PMID"
    PURL = "PURL"
    UPC = "UPC"
    URL = "URL"
    URN = "URN"


class DataciteRelationType(Enum):
    """Ways another object is related to the object describe by our DataCite record"""
    IsCitedBy = "IsCitedBy"
    Cites = "Cites"
    IsSupplementTo = "IsSupplementTo"
    IsSupplementedBy = "IsSupplementedBy"
    IsContinuedBy = "IsContinuedBy"
    Continues = "Continues"
    IsNewVersionOf = "IsNewVersionOf"
    IsPreviousVersionOf = "IsPreviousVersionOf"
    IsPartOf = "IsPartOf"
    HasPart = "HasPart"
    IsReferencedBy = "IsReferencedBy"
    References = "References"
    IsDocumentedBy = "IsDocumentedBy"
    Documents = "Documents"
    IsCompiledBy = "IsCompiledBy"
    Compiles = "Compiles"
    IsVariantFormOf = "IsVariantFormOf"
    IsOriginalFormOf = "IsOriginalFormOf"
    IsIdenticalTo = "IsIdenticalTo"
    HasMetadata = "HasMetadata"
    IsMetadataFor = "IsMetadataFor"
    Reviews = "Reviews"
    IsReviewedBy = "IsReviewedBy"
    IsDerivedFrom = "IsDerivedFrom"
    IsSourceOf = "IsSourceOf"


class DataciteRelatedIdentifier(BaseModel):
    """Identifier of a related resource. Use this property to indicate subsets of properties, as appropriate."""
    relatedIdentifier: str
    relatedIdentifierType: DataciteRelatedIdentifierType
    relatedMetadataScheme: str
    schemeURI: Optional[AnyUrl] = None


class DataciteDescription(BaseModel):
    """Additional information that does not fit in any of the other categories.

    May be used for technical information. It is a best practice to supply a description."""

    class DataciteDescriptionType(Enum):
        abstract = "Abstract"
        methods = "Methods"
        seriesInformation = "SeriesInformation"
        tableOfContents = "TableOfContents"
        technicalInfo = "TechnicalInfo"
        other = "Other"

    descriptionType: DataciteDescriptionType
    description: str


class DataciteRights(BaseModel):
    """"Any rights information for this resource. Provide a rights management statement for the resource or reference
     a service providing such information. Include embargo information if applicable. Use the complete title of a
     license and include version information if applicable."""

    rightsURI: Optional[str] = None
    rights: Optional[str] = None


class DataciteFundingReference(BaseModel):
    """Information about financial support (funding) for the resource being registered."""

    class FunderIdentifier(BaseModel):
        """Uniquely identifies a funding entity, according to various types."""

        class FunderIdentifierType(Enum):
            ISNI = "ISNI"
            GRID = "GRID"
            CrossRef = "Crossref Funder ID"
            Other = "Other"

        funderIdentifier: str
        funderIdentifierType: FunderIdentifierType

    class AwardNumber(BaseModel):
        awardNumber: str
        awardURI: Optional[AnyUrl] = None

    funderName: str
    funderIdentifier: Optional[FunderIdentifier] = None
    awardNumber: Optional[AwardNumber] = None
    awardTitle: Optional[str] = None


class Datacite(BaseModel):
    identifier: DataciteIdentifier = Field(default_factory=DataciteIdentifier)
    creators: List[DataciteCreator] = Field(default_factory=list)
    titles: List[DataciteTitle] = Field(default_factory=list)
    publisher: str = "DLHub"
    publicationYear: str = ""
    subjects: List[DataciteSubject] = Field(default_factory=list)
    resourceType: Optional[DataciteResourceType] = None
    contributors: List[DataciteContributor] = Field(default_factory=list)
    descriptions: List[DataciteDescription] = Field(default_factory=list)
    language: Optional[str] = None
    alternateIdentifiers: List[DataciteAlternateIdentifier] = Field(default_factory=list)
    relatedIdentifiers: List[DataciteRelatedIdentifier] = Field(default_factory=list)
    rightsList: List[DataciteRights] = Field(default_factory=list)
    fundingReferences: List[DataciteFundingReference] = Field(default_factory=list)
    version: Optional[str] = None

    __datacite_version: Optional[str] = "4.3"

    def set_title(self, title: str) -> 'Datacite':
        """Set the title for the artefact

        Use this method if you only have one title for the object

        Args:
            title - Desired title
        """
        self.titles = [DataciteTitle(title=title)]
        return self

    def set_creators(self, authors: List[str], affiliations: List[Sequence[str]] = ()) -> 'Datacite':
        """Add authors to an object

        Args:
            authors ([string]): List of authors for the dataset.
                In format: "<Family Name>, <Given Name>"
            affiliations ([[string]]): List of affiliations for each author.
        """
        for author, aff in zip_longest(authors, affiliations, fillvalue=[]):
            # Get the authors
            temp = author.split(",")
            family = temp[0].strip()
            given = temp[1].strip()

            # Add them to the list
            self.creators.append(DataciteCreator(
                givenName=given,
                familyName=family,
                affiliations=aff
            ))
        return self

    def set_abstract(self, abstract: str):
        """Define an abstract for this object

        Args:
            abstract: Text of abstract
        """

        abs_block = DataciteDescription(description=abstract, descriptionType='Abstract')
        self._add_or_replace_description(abs_block)
        return self

    def _add_or_replace_description(self, block: DataciteDescription):
        """Add a description or replace the text of an existing description of the same type

        Ensures no more than one of a certain type exists in the descriptions list

        Args:
            block: Description text and type block
        """
        # Delete block of same type if it already exists
        self.descriptions = [
            x for x in self.descriptions
            if x.descriptionType != block.descriptionType
        ]
        # Add the new block
        self.descriptions.append(block)

    def set_methods(self, methods: str):
        """Define the methods used to create this object

        Args:
            methods: Text of methods section
        """
        block = DataciteDescription(description=methods, descriptionType='Methods')
        self._add_or_replace_description(block)
        return self

    def set_identifier(self, doi: str) -> 'Datacite':
        """Set the identifier for this object

        Args:
            doi: Identifier for the object. Must be a DOI
        """
        self.identifier = DataciteIdentifier(identifier=doi)
        return self

    def add_rights(self, uri: Optional[str] = None, rights: Optional[str] = None) -> 'Datacite':
        """Any rights information for this resource. Provide a rights management statement for the
        resource or reference a service providing such information. Include embargo information if
        applicable. Use the complete title of a license and include
        version information if applicable.

        Args:
            uri (string): URI of the rights
            rights (string): Description of the rights
        """
        if uri is None and rights is None:
            raise ValueError('You must defined either a URI or the rights')

        # Make the rights
        new_rights = {}
        if uri is not None:
            new_rights['rightsURI'] = uri
        if new_rights is not None:
            new_rights['rights'] = rights

        # Add it to the list
        self.rightsList.append(DataciteRights(**new_rights))
        return self

    def add_funding_reference(self, name: str, identifier: Optional[str] = None,
                              identifier_type: Optional[str] = None,
                              award_number: Optional[str] = None,
                              award_title: Optional[str] = None,
                              award_uri: Optional[str] = None) -> 'Datacite':
        """Add a funding source to the list of resources

        Args:
            name: Name of funding provider
            identifier: Identifier (e.g., ISNI) of the funder
            identifier_type (string): Type of the identifier (ISNI, GRID, Crossref Funder ID, Other)
            award_number (string): Code assigned by the funder
            award_title (string): Title of the award
            award_uri (string): URI of the award
        """

        # Format in datacite format
        funder = {'funderName': name}
        if identifier is not None:
            funder['funderIdentifier'] = {'funderIdentifier': identifier,
                                          'funderIdentifierType': identifier_type}
        if award_number is not None:
            funder['awardNumber'] = {'awardNumber': award_number}
            if award_uri is not None:
                funder['awardNumber']['awardURI'] = award_uri
        if award_title is not None:
            funder['awardTitle'] = award_title

        # Append to the total list
        self.fundingReferences.append(DataciteFundingReference(**funder))
        return self

    def add_alternate_identifier(self, identifier: str, identifier_type: str) -> 'Datacite':
        """Add an identifier of this object in another service

        Args:
            identifier (string): Identifier
            identifier_type (string): Identifier type
        """

        self.alternateIdentifiers.append(DataciteAlternateIdentifier(
            alternateIdentifier=identifier,
            alternateIdentifierType=identifier_type
        ))
        return self

    def add_related_identifier(self, identifier: str, identifier_type: Union[str, DataciteRelatedIdentifierType],
                               relation_type: Union[str, DataciteRelationType]) -> 'Datacite':
        """Add an identifier of an artifact that is related to this resource (e.g., a paper
        that describes a dataset).

        You must define both the identifier and how it relates to this resource. The possible types
        of relations are listed in the
        `documentation for datacite
        <https://schema.datacite.org/meta/kernel-4.1/doc/DataCite-MetadataKernel_v4.1.pdf>_`
        on Page 25.

        Args:
            identifier: Identifier
            identifier_type: Identifier type
            relation_type: Relation between this identifier and the object you are describing
        """

        self.relatedIdentifiers.append(DataciteRelatedIdentifier(
            relatedIdentifier=identifier,
            relatedIdentifierType=identifier_type,
            relationType=relation_type
        ))
        return self

    def to_json(self, exclude_unset: bool = True) -> str:
        return self.json(exclude_unset=exclude_unset)

    def to_xml(self):
        raise NotImplementedError()

    def describe(self) -> str:
        return "Datacite version: {}".format(self.__datacite_version)

    @classmethod
    def from_codemeta(cls, metadata: dict) -> 'Datacite':
        """Generate Datacite metadata from a Codemeta

        Args:
            metadata: Codemeta-format metadata
        Returns:
            Datacite metadata object
        """

        # Taken from: https://github.com/caltechlibrary/ames/blob
        #             /920c028f7fcceaa1baf3a645a37de587dd2dbc41/matchers/caltechdata.py#L61

        datacite = {}
        if 'author' in metadata:
            creators = []
            for a in metadata['author']:
                cre = {}
                cre['creatorName'] = a['familyName'] + ', ' + a['givenName']
                cre['familyName'] = a['familyName']
                cre['givenName'] = a['givenName']
                if '@id' in a:
                    idv = a['@id']
                    split = idv.split('/')
                    idn = split[-1]
                    cre['nameIdentifiers'] = [{
                        'nameIdentifier': idn, 'nameIdentifierScheme': 'ORCID',
                        'schemeURI': 'http://orcid.org'}]
                    # Should check for type and remove hard code URI
                if 'affiliation' in a:
                    cre['affiliations'] = [a['affiliation']]
                    # Should check if can support multiple affiliations
                creators.append(cre)
            datacite['creators'] = creators
        if 'name' in metadata:
            datacite['titles'] = [{'title': metadata['name']}]
        if 'license' in metadata:
            # Assuming uri to name conversion, not optimal
            uri = metadata['license']
            name = uri.split('/')[-1]
            datacite['rightsList'] = [{'rights': name, 'rightsURI': uri}]
        if 'keywords' in metadata:
            sub = []
            for k in metadata['keywords']:
                sub.append({"subject": k})
            datacite['subjects'] = sub
        if 'funder' in metadata:
            # Kind of brittle due to limitations in codemeta
            fund_list = []
            grant_info = ''
            if 'funding' in metadata:
                grant_info = metadata['funding'].split(',')
            if isinstance(metadata['funder'], list):
                count = 0
                for funder in metadata['funder']:
                    entry = {'funderName': funder['name']}
                    if '@id' in funder:
                        element = {}
                        element['funderIdentifier'] = funder['@id']
                        element['funderIdentifierType'] = 'Crossref Funder ID'
                        entry['funderIdentifier'] = element
                    if grant_info != '':
                        split = grant_info[count].split(';')
                        entry['awardNumber'] = {'awardNumber': split[0]}
                        if len(split) > 1:
                            entry['awardTitle'] = split[1]
                    count = count + 1
            else:
                funder = metadata['funder']
                entry = {'funderName': funder['name']}
                if '@id' in funder:
                    element = {}
                    element['funderIdentifier'] = funder['@id']
                    element['funderIdentifierType'] = 'Crossref Funder ID'
                    entry['funderIdentifier'] = element
                if grant_info != '':
                    split = grant_info[0].split(';')
                    entry['awardNumber'] = {'awardNumber': split[0]}
                    if len(split) > 1:
                        entry['awardTitle'] = split[1]
                fund_list.append(entry)

            datacite['fundingReferences'] = fund_list
        return Datacite.parse_obj(datacite)



from typing import Dict, Optional, List, Any, Union

from pydantic import BaseModel, Field

from dlhub_sdk.models import BaseMetadataModel


class ArgumentTypeMetadata(BaseModel):
    """Description of an input argument"""

    type: str = Field(None, help="Type of the argument")
    description: Optional[str] = Field(None, help="Description of the argument")
    shape: Optional[List[Union[None, int]]] = None
    python_type: Optional[str] = None
    item_type: Optional['ArgumentTypeMetadata'] = None
    element_types: Optional[List['ArgumentTypeMetadata']] = None
    properties: Optional[Dict[str, 'ArgumentTypeMetadata']] = None


ArgumentTypeMetadata.update_forward_refs()


class MethodMetadata(BaseModel):
    """Metadata that describes each method"""

    input: ArgumentTypeMetadata = Field(..., help="Description of the method inputs")
    output: ArgumentTypeMetadata = Field(..., help="Description of the method outputs")
    parameters: Dict[str, Any] = Field(default_factory=dict, help="Description of method runtime parameters")
    method_details: Dict = Field(default_factory=dict, help="Options used to construct the method in DLHub.")


class ServableMetadata(BaseModel):
    """Metadata for servable objects.

    Captures the information that describe how to run a servable."""

    type: Optional[str] = Field(None, help="Type of the servable. Meant to be human readable")
    shim: Optional[str] = Field(None, help="Name of the home_run shim used to run a servable.")
    model_type: Optional[str] = Field(None, help="Simple description of the type of a machine learning model")
    model_summary: Optional[str] = Field(None, help="Longer-form description of a model.")
    methods: Dict[str, MethodMetadata] = Field(default_factory=dict,
                                               help="Description of each method for the servable")
    options: Optional[Dict] = Field(default_factory=dict, help="Settings used to construct the servable")

    class Config:
        extra = 'allow'


class BaseServableModel(BaseMetadataModel):
    """Base class for servables. Holds the metadata for the object and how to create and run the servable object."""

    servable: ServableMetadata = Field(ServableMetadata,
                                       help="Metadata describing how to construct and run a servable")

    def __init__(self):
        super(BaseServableModel, self).__init__()

        # Add the resource type
        #  I chose "InteractiveResource" as the point of DLHub is to provide
        #   web servies for interacting with these models ("query/response portals" are
        #   defined as "InteractiveResources") rather than downloading the source code
        #   (which would fit the definition of software)
        self.datacite.resourceType = {'resourceTypeGeneral': 'InteractiveResource'}

        # Define artifact type
        self.dlhub.type = 'servable'

        # Initialize the model running-information
        self.servable = ServableMetadata(shim=self._get_handler(), type=self._get_type())

    def _get_handler(self):
        """Generate the name of the servable class that DLHub will use to read this metadata

        Returns:
            (string) path to the python class (e.g., "python.PythonStaticMethod")
        """
        raise NotImplementedError()

    def _get_type(self):
        """Get a human-friendly name for this type of servable

        Returns:
            (string): Human-friendly name of an object
        """
        raise NotImplementedError()

    def register_function(self, name, inputs, outputs, parameters=None, method_details=None):
        """Registers a new function to this servable

        See :code:`compose_argument_type` utility function for how to define the inputs
        and outputs to this function.

        Args:
            name (string): Name of the function (e.g., "run")
            inputs (dict): Description of inputs to the function
            outputs (dict): Description of the outputs of the function
            parameters (dict): Any additional parameters for the function and their default values
            method_details (dict): Any options used when constructing a shim to run this function.
        """

        # Check defaults
        if method_details is None:
            method_details = {}
        if parameters is None:
            parameters = {}

        # Add the function
        self.servable.methods[name] = MethodMetadata.parse_obj({
            'input': inputs,
            'output': outputs,
            'parameters': parameters,
            'method_details': method_details
        })

        return self



import logging

try:
    import keras as keras_keras
except ImportError:
    keras_keras = None
try:
    from tensorflow import keras as tf_keras
except ImportError:
    tf_keras = None

from dlhub_sdk.models.servables.python import BasePythonServableModel
from dlhub_sdk.models.servables import ArgumentTypeMetadata
from dlhub_sdk.utils.types import compose_argument_block

logger = logging.getLogger(__name__)
_summary_limit = 10000


def _detect_backend(keras, output):
    """Add the backend

    Args:
        output (KerasModel): Current description of Keras model, will be modified
    """

    # Determine the name of the object
    my_backend = keras.backend.backend().lower()

    # Add it as a requirement
    output.add_requirement(my_backend, 'detect')


class KerasModel(BasePythonServableModel):
    """Servable based on a Keras Model object.

    Assumes that the model has been saved to an hdf5 file"""

    @classmethod
    def create_model(cls, model_path, output_names=None, arch_path=None,
                     custom_objects=None, force_tf_keras: bool = False) -> 'KerasModel':
        """Initialize a Keras model.

        Args:
            model_path (string): Path to the hd5 file that contains the weights and, optionally,
                the architecture
            output_names ([string] or [[string]]): Names of output classes.
            arch_path (string): Path to the hd5 model containing the architecture, if not
                available in the file at :code:`model_path`.
            custom_objects (dict): Map of layer names to custom layers. See
                `Keras Documentation
                <https://www.tensorflow.org/api_docs/python/tf/keras/models/load_model>`_
                for more details.
            force_tf_keras (bool): Force the use of TF.Keras even if keras is installed
       """
        output: KerasModel = super().create_model('predict')
        if force_tf_keras:
            if tf_keras is None:
                raise ValueError('You forced tf_keras but do not have tensorflow.keras')
            keras = tf_keras
            use_tf_keras = True
        elif keras_keras is not None:
            # Use old keras by default, as users may have gone out of their way to install it
            keras = keras_keras
            use_tf_keras = False
            if tf_keras is not None:
                logging.warning('Model publication will use standalone keras, yet you have tf.keras installed. '
                                'If you want your model to use tf.keras, use ``force_tf_keras=True``.')
        elif tf_keras is not None:
            keras = tf_keras
            use_tf_keras = True
        else:
            raise ValueError('You do not have any version of keras installed.')

        # Add model as a file to be sent
        output.add_file(model_path, 'model')
        if arch_path is not None:
            output.add_file(arch_path, 'arch')

        # Store the list of custom objects
        if custom_objects is not None:
            for k, v in custom_objects.items():
                output.add_custom_object(k, v)

        # Get the model details
        if arch_path is None:
            model = keras.models.load_model(model_path, custom_objects=custom_objects)
        else:
            if arch_path.endswith('.h5') or arch_path.endswith('.hdf') \
                    or arch_path.endswith('.hdf5') or arch_path.endswith('.hd5'):
                model = keras.models.load_model(arch_path,
                                                custom_objects=custom_objects, compile=False)
            elif arch_path.endswith('.json'):
                with open(arch_path) as fp:
                    json_string = fp.read()
                model = keras.models.model_from_json(json_string, custom_objects=custom_objects)
            elif arch_path.endswith('.yml') or arch_path.endswith('.yaml'):
                with open(arch_path) as fp:
                    yaml_string = fp.read()
                model = keras.models.model_from_yaml(yaml_string, custom_objects=custom_objects)
            else:
                raise ValueError('File type for architecture not recognized')
            model.load_weights(model_path)

        # Get the inputs of the model
        output.servable.methods['run'].input = output.format_layer_spec(model.input_shape)
        output.servable.methods['run'].output = output.format_layer_spec(model.output_shape)
        if output_names is not None:
            output.servable.methods['run'].method_details['classes'] = output_names

        # Get a full description of the model. Limit summary to _summary_limit in length
        _summary_tmp = []
        model.summary(print_fn=_summary_tmp.append)
        summary = "\n".join(_summary_tmp)
        summary = (summary[:_summary_limit] + '<<TRUNCATED>>') if len(summary) > _summary_limit else summary

        output.servable.model_summary = summary
        output.servable.model_type = 'Deep NN'

        # Add keras as a dependency
        keras_version = keras.__version__
        if not (keras_version.endswith("-tf") or use_tf_keras):
            output.add_requirement('keras', keras_version)
        output.add_requirement('h5py', 'detect')

        # Detect backend and get its version
        _detect_backend(keras, output)

        return output

    def format_layer_spec(self, layers) -> ArgumentTypeMetadata:
        """Make a description of a list of input or output layers

        Args:
            layers (tuple or [tuple]): Shape of the layers
        Return:
            (dict) Description of the inputs / outputs
        """
        if isinstance(layers, tuple):
            return ArgumentTypeMetadata.parse_obj(compose_argument_block("ndarray", "Tensor", shape=list(layers)))
        else:
            return ArgumentTypeMetadata.parse_obj(
                compose_argument_block("tuple", "Tuple of tensors",
                                       element_types=[self.format_layer_spec(i) for i in layers])
            )

    def add_custom_object(self, name, custom_object):
        """Add a custom layer to the model specification

        See `Keras FAQs
        <https://keras.io/getting-started/faq/#handling-custom-layers-or-other-custom-objects-in-saved-models>`
        for details.

        Args:
              name (string): Name of the custom object
              custom_object (class): Class of the custom object
        Return:
            self
        """

        # get the class name for the custom object
        object_name = custom_object.__name__
        module = custom_object.__module__

        # Add the layer to the model definition
        if 'custom_objects' not in self.servable.options:
            self.servable.options['custom_objects'] = {}
        self.servable.options['custom_objects'][name] = '{}.{}'.format(module, object_name)

    def _get_handler(self):
        return "keras.KerasServable"

    def _get_type(self):
        return "Keras Model"
